@@ -7,6 +7,7 @@ use self_cell::self_cell;
 
 use crate::environment::CompiledTemplate;
 use crate::error::{Error, ErrorKind};
+use crate::utils::BTreeMapKeysDebug;
 use crate::value::RcType;
 
 /// Utility for dynamic template loading.
@@ -40,13 +41,13 @@ pub struct Source {
 
 impl fmt::Debug for Source {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        fmt::Debug::fmt(&self.templates, f)
+        fmt::Debug::fmt(&BTreeMapKeysDebug(&self.templates), f)
     }
 }
 
 self_cell! {
     struct LoadedTemplate {
-        owner: String,
+        owner: (String, String),
         #[covariant]
         dependent: CompiledTemplate,
     }
@@ -76,7 +77,8 @@ impl Source {
     ) -> Result<(), Error> {
         let source = source.into();
         let name = name.into();
-        let tmpl = LoadedTemplate::try_new(source, |source| -> Result<_, Error> {
+        let owner = (name.clone(), source);
+        let tmpl = LoadedTemplate::try_new(owner, |(name, source)| -> Result<_, Error> {
             CompiledTemplate::from_name_and_source(name.as_str(), source)
         })?;
         self.templates.insert(name, RcType::new(tmpl));
