@@ -6,11 +6,11 @@
 //! is indeed an odd number.
 //!
 //! MiniJinja comes with some built-in test functions that are listed below. To
-//! create a custom test write a function that takes at least an
-//! [`&Environment`](crate::Environment) and value argument and returns a boolean
+//! create a custom test write a function that takes at least a
+//! [`&State`](crate::State) and value argument and returns a boolean
 //! result, then register it with [`add_filter`](crate::Environment::add_test).
 //!
-//! ## Custom Tests
+//! # Custom Tests
 //!
 //! A custom test function is just a simple function which accepts inputs as
 //! parameters and then returns a bool wrapped in a result. For instance the
@@ -18,9 +18,9 @@
 //! it's lowercase:
 //!
 //! ```
-//! # use minijinja::{Environment, Error};
+//! # use minijinja::{State, Environment, Error};
 //! # let mut env = Environment::new();
-//! fn is_lowercase(env: &Environment, value: String) -> Result<bool, Error> {
+//! fn is_lowercase(_state: &State, value: String) -> Result<bool, Error> {
 //!    Ok(value.chars().all(|x| x.is_lowercase()))
 //! }
 //!
@@ -31,12 +31,11 @@
 //! [`FunctionArgs`](crate::value::FunctionArgs) trait.
 use std::collections::BTreeMap;
 
-use crate::environment::Environment;
 use crate::error::Error;
 use crate::value::{ArgType, FunctionArgs, RcType, Value};
+use crate::vm::State;
 
-type TestFunc =
-    dyn Fn(&Environment, Value, Vec<Value>) -> Result<bool, Error> + Sync + Send + 'static;
+type TestFunc = dyn Fn(&State, Value, Vec<Value>) -> Result<bool, Error> + Sync + Send + 'static;
 
 #[derive(Clone)]
 pub(crate) struct BoxedTest(RcType<TestFunc>);
@@ -44,19 +43,19 @@ pub(crate) struct BoxedTest(RcType<TestFunc>);
 /// A utility trait that represents filters.
 pub trait Test<V = Value, Args = Vec<Value>>: Send + Sync + 'static {
     /// Performs a test to value with the given arguments.
-    fn perform(&self, env: &Environment, value: V, args: Args) -> Result<bool, Error>;
+    fn perform(&self, state: &State, value: V, args: Args) -> Result<bool, Error>;
 }
 
 macro_rules! tuple_impls {
     ( $( $name:ident )* ) => {
         impl<Func, V, $($name),*> Test<V, ($($name,)*)> for Func
         where
-            Func: Fn(&Environment, V, $($name),*) -> Result<bool, Error> + Send + Sync + 'static
+            Func: Fn(&State, V, $($name),*) -> Result<bool, Error> + Send + Sync + 'static
         {
-            fn perform(&self, env: &Environment, value: V, args: ($($name,)*)) -> Result<bool, Error> {
+            fn perform(&self, state: &State, value: V, args: ($($name,)*)) -> Result<bool, Error> {
                 #[allow(non_snake_case)]
                 let ($($name,)*) = args;
-                (self)(env, value, $($name,)*)
+                (self)(state, value, $($name,)*)
             }
         }
     };
@@ -77,9 +76,9 @@ impl BoxedTest {
         Args: FunctionArgs,
     {
         BoxedTest(RcType::new(
-            move |env, value, args| -> Result<bool, Error> {
+            move |state, value, args| -> Result<bool, Error> {
                 f.perform(
-                    env,
+                    state,
                     ArgType::from_value(Some(value))?,
                     FunctionArgs::from_values(args)?,
                 )
@@ -88,13 +87,8 @@ impl BoxedTest {
     }
 
     /// Applies the filter to a value and argument.
-    pub fn perform(
-        &self,
-        env: &Environment,
-        value: Value,
-        args: Vec<Value>,
-    ) -> Result<bool, Error> {
-        (self.0)(env, value, args)
+    pub fn perform(&self, state: &State, value: Value, args: Vec<Value>) -> Result<bool, Error> {
+        (self.0)(state, value, args)
     }
 }
 
@@ -126,7 +120,7 @@ mod builtins {
 
     /// Checks if a value is odd.
     #[cfg_attr(docsrs, doc(cfg(feature = "builtin_tests")))]
-    pub fn is_odd(_env: &Environment, v: Value) -> Result<bool, Error> {
+    pub fn is_odd(_state: &State, v: Value) -> Result<bool, Error> {
         Ok(v.as_primitive()
             .and_then(|x| x.as_i128())
             .map_or(false, |x| x % 2 != 0))
@@ -134,7 +128,7 @@ mod builtins {
 
     /// Checks if a value is even.
     #[cfg_attr(docsrs, doc(cfg(feature = "builtin_tests")))]
-    pub fn is_even(_env: &Environment, v: Value) -> Result<bool, Error> {
+    pub fn is_even(_state: &State, v: Value) -> Result<bool, Error> {
         Ok(v.as_primitive()
             .and_then(|x| x.as_i128())
             .map_or(false, |x| x % 2 == 0))
@@ -142,62 +136,69 @@ mod builtins {
 
     /// Checks if a value is undefined.
     #[cfg_attr(docsrs, doc(cfg(feature = "builtin_tests")))]
-    pub fn is_undefined(_env: &Environment, v: Value) -> Result<bool, Error> {
+    pub fn is_undefined(_state: &State, v: Value) -> Result<bool, Error> {
         Ok(v.is_undefined())
     }
 
     /// Checks if a value is defined.
     #[cfg_attr(docsrs, doc(cfg(feature = "builtin_tests")))]
-    pub fn is_defined(_env: &Environment, v: Value) -> Result<bool, Error> {
+    pub fn is_defined(_state: &State, v: Value) -> Result<bool, Error> {
         Ok(!v.is_undefined())
     }
 
     /// Checks if this value is a number.
     #[cfg_attr(docsrs, doc(cfg(feature = "builtin_tests")))]
-    pub fn is_number(_env: &Environment, v: Value) -> Result<bool, Error> {
+    pub fn is_number(_state: &State, v: Value) -> Result<bool, Error> {
         Ok(matches!(v.kind(), ValueKind::Number))
     }
 
     /// Checks if this value is a string.
     #[cfg_attr(docsrs, doc(cfg(feature = "builtin_tests")))]
-    pub fn is_string(_env: &Environment, v: Value) -> Result<bool, Error> {
+    pub fn is_string(_state: &State, v: Value) -> Result<bool, Error> {
         Ok(matches!(v.kind(), ValueKind::String))
     }
 
     /// Checks if this value is a sequence
     #[cfg_attr(docsrs, doc(cfg(feature = "builtin_tests")))]
-    pub fn is_sequence(_env: &Environment, v: Value) -> Result<bool, Error> {
+    pub fn is_sequence(_state: &State, v: Value) -> Result<bool, Error> {
         Ok(matches!(v.kind(), ValueKind::Seq))
     }
 
     /// Checks if this value is a mapping
     #[cfg_attr(docsrs, doc(cfg(feature = "builtin_tests")))]
-    pub fn is_mapping(_env: &Environment, v: Value) -> Result<bool, Error> {
+    pub fn is_mapping(_state: &State, v: Value) -> Result<bool, Error> {
         Ok(matches!(v.kind(), ValueKind::Map))
     }
 
     /// Checks if the value is starting with a string.
     #[cfg_attr(docsrs, doc(cfg(feature = "builtin_tests")))]
-    pub fn is_startingwith(_env: &Environment, v: String, other: String) -> Result<bool, Error> {
+    pub fn is_startingwith(_state: &State, v: String, other: String) -> Result<bool, Error> {
         Ok(v.starts_with(&other))
     }
 
     /// Checks if the value is ending with a string.
     #[cfg_attr(docsrs, doc(cfg(feature = "builtin_tests")))]
-    pub fn is_endingwith(_env: &Environment, v: String, other: String) -> Result<bool, Error> {
+    pub fn is_endingwith(_state: &State, v: String, other: String) -> Result<bool, Error> {
         Ok(v.ends_with(&other))
     }
 
     #[test]
     fn test_basics() {
-        fn test(_: &Environment, a: u32, b: u32) -> Result<bool, Error> {
+        fn test(_: &State, a: u32, b: u32) -> Result<bool, Error> {
             Ok(a == b)
         }
 
-        let env = Environment::new();
+        let env = crate::Environment::new();
+        let state = State {
+            env: &env,
+            ctx: crate::vm::Context::default(),
+            auto_escape: crate::AutoEscape::None,
+            current_block: None,
+            name: "<unknown>",
+        };
         let bx = BoxedTest::new(test);
         assert!(bx
-            .perform(&env, Value::from(23), vec![Value::from(23)])
+            .perform(&state, Value::from(23), vec![Value::from(23)])
             .unwrap());
     }
 }
