@@ -76,10 +76,10 @@ use std::sync::atomic::{self, AtomicBool, AtomicUsize};
 
 use serde::ser::{self, Serialize, Serializer};
 
-use crate::environment::Environment;
 use crate::error::{Error, ErrorKind};
 use crate::key::{Key, KeySerializer};
 use crate::utils::matches;
+use crate::vm::State;
 
 #[cfg(feature = "sync")]
 pub(crate) type RcType<T> = std::sync::Arc<T>;
@@ -175,6 +175,24 @@ pub enum ValueKind {
     Struct,
 }
 
+impl fmt::Display for ValueKind {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        let ty = match *self {
+            ValueKind::Undefined => "undefined",
+            ValueKind::None => "none",
+            ValueKind::Bool => "bool",
+            ValueKind::Number => "number",
+            ValueKind::Char => "char",
+            ValueKind::String => "string",
+            ValueKind::Bytes => "bytes",
+            ValueKind::Seq => "sequence",
+            ValueKind::Map => "map",
+            ValueKind::Struct => "struct",
+        };
+        write!(f, "{}", ty)
+    }
+}
+
 #[derive(Clone)]
 enum Repr {
     Undefined,
@@ -549,7 +567,12 @@ macro_rules! math_binop {
             do_it(lhs, rhs).ok_or_else(|| {
                 Error::new(
                     ErrorKind::ImpossibleOperation,
-                    concat!("tried to use ", stringify!($float), " operator on unsupported types")
+                    format!(
+                        "tried to use {} operator on unsupported types {} and {}",
+                        stringify!($float),
+                        lhs.kind(),
+                        rhs.kind()
+                    )
                 )
             })
         }
@@ -1017,10 +1040,10 @@ impl Value {
     }
 
     /// Calls the value directly.
-    pub(crate) fn call(&self, env: &Environment, args: Vec<Value>) -> Result<Value, Error> {
+    pub(crate) fn call(&self, state: &State, args: Vec<Value>) -> Result<Value, Error> {
         if let Repr::Shared(ref cplx) = self.0 {
             if let Shared::Dynamic(ref dy) = **cplx {
-                return dy.call(env, args);
+                return dy.call(state, args);
             }
         }
         Err(Error::new(
@@ -1032,13 +1055,13 @@ impl Value {
     /// Calls a method on the value.
     pub(crate) fn call_method(
         &self,
-        env: &Environment,
+        state: &State,
         name: &str,
         args: Vec<Value>,
     ) -> Result<Value, Error> {
         if let Repr::Shared(ref cplx) = self.0 {
             if let Shared::Dynamic(ref dy) = **cplx {
-                return dy.call_method(env, name, args);
+                return dy.call_method(state, name, args);
             }
         }
         Err(Error::new(
@@ -1615,16 +1638,20 @@ impl<'a> fmt::Debug for ValueIterator {
 /// value to pass.  Here a simple motivating example:
 ///
 /// ```
-/// use minijinja::{Environment, value::Single};
+/// use minijinja::{Environment, context};
 ///
 /// let mut env = Environment::new();
 /// env.add_template("simple", "Hello {{ name }}!").unwrap();
 /// let tmpl = env.get_template("simple").unwrap();
-/// let rv = tmpl.render(Single("name", "Peter")).unwrap();
+/// let rv = tmpl.render(context!(name => "Peter")).unwrap();
 /// ```
+///
+/// This type is deprecated and has been replaced with the [`context`] macro.
 #[derive(Debug, Clone)]
+#[deprecated(since = "0.8.0", note = "replaced by the more generic context! macro")]
 pub struct Single<'a, V: Serialize + ?Sized>(pub &'a str, pub V);
 
+#[allow(deprecated)]
 impl<'a, V: Serialize + ?Sized> Serialize for Single<'a, V> {
     fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
     where
@@ -1698,8 +1725,8 @@ pub trait Object: fmt::Display + fmt::Debug + Any + Sync + Send {
     ///
     /// It's the responsibility of the implementer to ensure that an
     /// error is generated if an invalid method is invoked.
-    fn call_method(&self, env: &Environment, name: &str, args: Vec<Value>) -> Result<Value, Error> {
-        let _env = env;
+    fn call_method(&self, state: &State, name: &str, args: Vec<Value>) -> Result<Value, Error> {
+        let _state = state;
         let _args = args;
         Err(Error::new(
             ErrorKind::ImpossibleOperation,
@@ -1711,8 +1738,8 @@ pub trait Object: fmt::Display + fmt::Debug + Any + Sync + Send {
     ///
     /// The default implementation just generates an error that the object
     /// cannot be invoked.
-    fn call(&self, env: &Environment, args: Vec<Value>) -> Result<Value, Error> {
-        let _env = env;
+    fn call(&self, state: &State, args: Vec<Value>) -> Result<Value, Error> {
+        let _state = state;
         let _args = args;
         Err(Error::new(
             ErrorKind::ImpossibleOperation,
@@ -1734,7 +1761,7 @@ fn test_adding() {
     let err = add(&value!("a"), &value!(42)).unwrap_err();
     assert_eq!(
         err.to_string(),
-        "impossible operation: tried to use + operator on unsupported types"
+        "impossible operation: tried to use + operator on unsupported types string and number"
     );
 
     assert_eq!(add(&value!(1), &value!(2)), Ok(value!(3)));
