@@ -3,8 +3,6 @@ use std::convert::TryFrom;
 use std::fmt::{self, Write};
 use std::sync::atomic::{AtomicUsize, Ordering};
 
-use serde::Serialize;
-
 use crate::environment::Environment;
 use crate::error::{Error, ErrorKind};
 use crate::instructions::{Instruction, Instructions};
@@ -13,12 +11,21 @@ use crate::utils::matches;
 use crate::value::{self, Object, Primitive, RcType, Value, ValueIterator};
 use crate::AutoEscape;
 
-#[derive(Debug)]
 pub struct LoopState {
     len: AtomicUsize,
     idx: AtomicUsize,
 }
 
+impl fmt::Debug for LoopState {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        let mut s = f.debug_struct("LoopState");
+        for attr in self.attributes() {
+            s.field(attr, &self.get_attr(attr).unwrap());
+        }
+        s.finish()
+    }
+}
+
 impl Object for LoopState {
     fn attributes(&self) -> &[&str] {
         &[
@@ -47,12 +54,7 @@ impl Object for LoopState {
         }
     }
 
-    fn call_method(
-        &self,
-        _env: &Environment,
-        name: &str,
-        args: Vec<Value>,
-    ) -> Result<Value, Error> {
+    fn call_method(&self, _state: &State, name: &str, args: Vec<Value>) -> Result<Value, Error> {
         if name == "cycle" {
             let idx = self.idx.load(Ordering::Relaxed);
             match args.get(idx % args.len()) {
@@ -70,7 +72,12 @@ impl Object for LoopState {
 
 impl fmt::Display for LoopState {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        write!(f, "loop")
+        write!(
+            f,
+            "<loop {}/{}>",
+            self.idx.load(Ordering::Relaxed),
+            self.len.load(Ordering::Relaxed)
+        )
     }
 }
 
@@ -82,24 +89,33 @@ pub struct Loop<'env> {
     controller: RcType<LoopState>,
 }
 
-#[derive(Debug)]
-pub enum Frame<'env, 'context> {
+pub enum Frame<'env, 'vm> {
     // This layer dispatches to another context
-    Chained {
-        base: &'context Context<'env, 'context>,
-    },
+    Chained { base: &'vm Context<'env, 'vm> },
     // this layer isolates
-    Isolate {
-        value: Value,
-    },
+    Isolate { value: Value },
     // this layer shadows another one
-    Merge {
-        value: Value,
-    },
+    Merge { value: Value },
     // this layer is a for loop
     Loop(Loop<'env>),
 }
 
+impl<'env, 'vm> fmt::Debug for Frame<'env, 'vm> {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        match self {
+            Self::Chained { base } => fmt::Debug::fmt(base, f),
+            Self::Isolate { value } => fmt::Debug::fmt(value, f),
+            Self::Merge { value } => fmt::Debug::fmt(value, f),
+            Self::Loop(l) => {
+                let mut m = f.debug_map();
+                m.entries(l.locals.iter());
+                m.entry(&"loop", &l.controller);
+                m.finish()
+            }
+        }
+    }
+}
+
 #[derive(Debug, Default)]
 pub struct Stack {
     values: Vec<Value>,
@@ -124,23 +140,23 @@ impl Stack {
 }
 
 #[derive(Default, Debug)]
-pub struct Context<'env, 'context> {
-    stack: Vec<Frame<'env, 'context>>,
+pub struct Context<'env, 'vm> {
+    stack: Vec<Frame<'env, 'vm>>,
 }
 
-impl<'env, 'context> Context<'env, 'context> {
+impl<'env, 'vm> Context<'env, 'vm> {
     /// Stores a variable in the context.
     pub fn store(&mut self, key: &'env str, value: Value) {
         self.current_loop().locals.insert(key, value);
     }
 
     /// Looks up a variable in the context.
-    pub fn lookup(&self, env: &Environment, key: &str) -> Option<Value> {
+    pub fn load(&self, env: &Environment, key: &str) -> Option<Value> {
         for ctx in self.stack.iter().rev() {
             let (lookup_base, cont) = match ctx {
                 // if we hit a chain frame we dispatch there and never
                 // recurse
-                Frame::Chained { base } => return base.lookup(env, key),
+                Frame::Chained { base } => return base.load(env, key),
                 Frame::Isolate { value } => (value, false),
                 Frame::Merge { value } => (value, true),
                 Frame::Loop(Loop {
@@ -175,7 +191,7 @@ impl<'env, 'context> Context<'env, 'context> {
     }
 
     /// Pushes a new layer.
-    pub fn push_frame(&mut self, layer: Frame<'env, 'context>) {
+    pub fn push_frame(&mut self, layer: Frame<'env, 'vm>) {
         self.stack.push(layer);
     }
 
@@ -198,6 +214,93 @@ impl<'env, 'context> Context<'env, 'context> {
     }
 }
 
+/// Provides access to the current execution state of the engine.
+///
+/// A read only reference is passed to filter functions and similar objects to
+/// allow limited interfacing with the engine.  The state is useful to look up
+/// information about the engine in filter, test or global functions.  It not
+/// only provides access to the template environment but also the context
+/// variables of the engine, the current auto escaping behavior as well as the
+/// auto escape flag.
+pub struct State<'vm, 'env> {
+    pub(crate) env: &'env Environment<'env>,
+    pub(crate) ctx: Context<'env, 'vm>,
+    pub(crate) name: &'env str,
+    pub(crate) current_block: Option<&'env str>,
+    pub(crate) auto_escape: AutoEscape,
+}
+
+impl<'vm, 'env> fmt::Debug for State<'vm, 'env> {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.debug_struct("State")
+            .field("ctx", &self.ctx)
+            .field("name", &self.name)
+            .field("current_block", &self.current_block)
+            .field("auto_escape", &self.auto_escape)
+            .field("env", &self.env)
+            .finish()
+    }
+}
+
+impl<'vm, 'env> State<'vm, 'env> {
+    /// Returns a reference to the current environment.
+    pub fn env(&self) -> &Environment<'env> {
+        self.env
+    }
+
+    /// Returns the name of the current template.
+    pub fn name(&self) -> &str {
+        self.name
+    }
+
+    /// Returns the current value of the auto escape flag.
+    pub fn auto_escape(&self) -> AutoEscape {
+        self.auto_escape
+    }
+
+    /// Returns the name of the innermost block.
+    pub fn current_block(&self) -> Option<&str> {
+        self.current_block
+    }
+
+    /// Looks up a variable by name in the context.
+    pub fn lookup(&self, name: &str) -> Option<Value> {
+        self.ctx.load(self.env(), name)
+    }
+
+    pub(crate) fn apply_filter(
+        &self,
+        name: &str,
+        value: Value,
+        args: Vec<Value>,
+    ) -> Result<Value, Error> {
+        if let Some(filter) = self.env().get_filter(name) {
+            filter.apply_to(self, value, args)
+        } else {
+            Err(Error::new(
+                ErrorKind::UnknownFilter,
+                format!("filter {} is unknown", name),
+            ))
+        }
+    }
+
+    pub(crate) fn perform_test(
+        &self,
+        name: &str,
+        value: Value,
+        args: Vec<Value>,
+    ) -> Result<bool, Error> {
+        if let Some(test) = self.env().get_test(name) {
+            test.perform(self, value, args)
+        } else {
+            Err(Error::new(
+                ErrorKind::UnknownTest,
+                format!("test {} is unknown", name),
+            ))
+        }
+    }
+}
+
 /// Helps to evaluate something.
 #[derive(Debug)]
 pub struct Vm<'env> {
@@ -211,54 +314,50 @@ impl<'env> Vm<'env> {
     }
 
     /// Evaluates the given inputs
-    pub fn eval<S: Serialize>(
+    pub fn eval(
         &self,
         instructions: &Instructions<'env>,
-        root: S,
+        root: Value,
         blocks: &BTreeMap<&'env str, Instructions<'env>>,
         initial_auto_escape: AutoEscape,
         output: &mut String,
     ) -> Result<Option<Value>, Error> {
-        let mut context = Context::default();
-        let root = Value::from_serializable(&root);
-        context.push_frame(Frame::Isolate { value: root });
+        let mut ctx = Context::default();
+        ctx.push_frame(Frame::Isolate { value: root });
         let mut referenced_blocks = BTreeMap::new();
         for (&name, instr) in blocks.iter() {
             referenced_blocks.insert(name, vec![instr]);
         }
-        let mut block_stack = vec![];
-        self.eval_context(
-            instructions,
-            &mut context,
-            &referenced_blocks,
-            &mut block_stack,
-            initial_auto_escape,
-            output,
-        )
+        let mut state = State {
+            env: self.env,
+            ctx,
+            auto_escape: initial_auto_escape,
+            current_block: None,
+            name: instructions.name(),
+        };
+        self.eval_state(&mut state, instructions, referenced_blocks, output)
     }
 
     /// This is the actual evaluation loop that works with a specific context.
-    fn eval_context(
+    fn eval_state(
         &self,
-        mut instructions: &'env Instructions<'env>,
-        context: &mut Context<'env, '_>,
-        blocks: &BTreeMap<&'env str, Vec<&'env Instructions<'env>>>,
-        block_stack: &mut Vec<&'env str>,
-        initial_auto_escape: AutoEscape,
+        state: &mut State<'_, 'env>,
+        mut instructions: &Instructions<'env>,
+        mut blocks: BTreeMap<&'env str, Vec<&'_ Instructions<'env>>>,
         output: &mut String,
     ) -> Result<Option<Value>, Error> {
-        let mut pc = 0;
+        let initial_auto_escape = state.auto_escape;
         let mut stack = Stack::default();
-        let mut blocks = blocks.clone();
-        let mut auto_escape = initial_auto_escape;
         let mut auto_escape_stack = vec![];
         let mut capture_stack = vec![];
+        let mut block_stack = vec![];
+        let mut pc = 0;
 
         macro_rules! bail {
             ($err:expr) => {{
                 let mut err = $err;
-                if let Some((filename, lineno)) = instructions.get_location(pc) {
-                    err.set_location(filename, lineno);
+                if let Some(lineno) = instructions.get_line(pc) {
+                    err.set_location(instructions.name(), lineno);
                 }
                 return Err(err);
             }};
@@ -305,7 +404,7 @@ impl<'env> Vm<'env> {
             () => {{
                 let captured = capture_stack.pop().unwrap();
                 // TODO: this should take the right auto escapine flag into account
-                stack.push(if !matches!(auto_escape, AutoEscape::None) {
+                stack.push(if !matches!(state.auto_escape, AutoEscape::None) {
                     Value::from_safe_string(captured)
                 } else {
                     Value::from(captured)
@@ -315,20 +414,24 @@ impl<'env> Vm<'env> {
 
         macro_rules! sub_eval {
             ($instructions:expr) => {{
-                sub_eval!($instructions, &blocks, block_stack, auto_escape);
-            }};
-            ($instructions:expr, $blocks:expr, $block_stack:expr, $auto_escape:expr) => {{
-                let mut sub_context = Context::default();
-                sub_context.push_frame(Frame::Chained { base: context });
-                let sub_vm = Vm::new(self.env);
-                sub_vm.eval_context(
+                sub_eval!(
                     $instructions,
-                    &mut sub_context,
-                    $blocks,
-                    $block_stack,
-                    $auto_escape,
-                    out!(),
-                )?;
+                    blocks.clone(),
+                    state.current_block,
+                    state.auto_escape
+                );
+            }};
+            ($instructions:expr, $blocks:expr, $current_block:expr, $auto_escape:expr) => {{
+                let mut sub_context = Context::default();
+                sub_context.push_frame(Frame::Chained { base: &state.ctx });
+                let mut sub_state = State {
+                    env: self.env,
+                    ctx: sub_context,
+                    auto_escape: $auto_escape,
+                    current_block: $current_block,
+                    name: $instructions.name(),
+                };
+                self.eval_state(&mut sub_state, $instructions, $blocks, out!())?;
             }};
         }
 
@@ -338,13 +441,13 @@ impl<'env> Vm<'env> {
                     write!(out!(), "{}", val).unwrap();
                 }
                 Instruction::Emit => {
-                    try_ctx!(self.env.finalize(&stack.pop(), auto_escape, out!()));
+                    try_ctx!(self.env.finalize(&stack.pop(), state.auto_escape, out!()));
                 }
                 Instruction::StoreLocal(name) => {
-                    context.store(name, stack.pop());
+                    state.ctx.store(name, stack.pop());
                 }
                 Instruction::Lookup(name) => {
-                    stack.push(context.lookup(self.env, name).unwrap_or(Value::UNDEFINED));
+                    stack.push(state.ctx.load(self.env, name).unwrap_or(Value::UNDEFINED));
                 }
                 Instruction::GetAttr(name) => {
                     let value = stack.pop();
@@ -435,16 +538,16 @@ impl<'env> Vm<'env> {
                 }
                 Instruction::PushContext => {
                     let value = stack.pop();
-                    context.push_frame(Frame::Merge { value });
+                    state.ctx.push_frame(Frame::Merge { value });
                 }
                 Instruction::PopFrame => {
-                    context.pop_frame();
+                    state.ctx.pop_frame();
                 }
                 Instruction::PushLoop(with_loop_var) => {
                     let iterable = stack.pop();
                     let iterator = iterable.iter();
                     let len = iterator.len();
-                    context.push_frame(Frame::Loop(Loop {
+                    state.ctx.push_frame(Frame::Loop(Loop {
                         locals: BTreeMap::new(),
                         iterator,
                         with_loop_var: *with_loop_var,
@@ -455,7 +558,7 @@ impl<'env> Vm<'env> {
                     }));
                 }
                 Instruction::Iterate(jump_target) => {
-                    let l = context.current_loop();
+                    let l = state.ctx.current_loop();
                     l.controller.idx.fetch_add(1, Ordering::Relaxed);
                     match l.iterator.next() {
                         Some(item) => {
@@ -494,7 +597,8 @@ impl<'env> Vm<'env> {
                     }
                 }
                 Instruction::CallBlock(name) => {
-                    block_stack.push(name);
+                    block_stack.push(state.current_block);
+                    state.current_block = Some(name);
                     if let Some(layers) = blocks.get(name) {
                         let instructions = layers.first().unwrap();
                         sub_eval!(instructions);
@@ -504,7 +608,7 @@ impl<'env> Vm<'env> {
                             "tried to invoke unknown block"
                         ));
                     }
-                    block_stack.pop();
+                    state.current_block = block_stack.pop().unwrap();
                 }
                 Instruction::LoadBlocks => {
                     let name = stack.pop();
@@ -528,37 +632,35 @@ impl<'env> Vm<'env> {
                     // execute the extended template's code instead.  From this
                     // there is no way back.
                     instructions = tmpl.instructions();
+                    state.name = instructions.name();
                     pc = 0;
                     continue;
                 }
                 Instruction::Include => {
                     let name = stack.pop();
-                    let tmpl = try_ctx!(name
-                        .as_str()
-                        .ok_or_else(|| {
-                            Error::new(
-                                ErrorKind::ImpossibleOperation,
-                                "template name was not a string",
-                            )
-                        })
-                        .and_then(|name| self.env.get_template(name)));
+                    let name = try_ctx!(name.as_str().ok_or_else(|| {
+                        Error::new(
+                            ErrorKind::ImpossibleOperation,
+                            "template name was not a string",
+                        )
+                    }));
+                    let tmpl = try_ctx!(self.env.get_template(name));
                     let instructions = tmpl.instructions();
                     let mut referenced_blocks = BTreeMap::new();
                     for (&name, instr) in tmpl.blocks().iter() {
                         referenced_blocks.insert(name, vec![instr]);
                     }
-                    let mut block_stack = Vec::new();
                     sub_eval!(
                         instructions,
-                        &referenced_blocks,
-                        &mut block_stack,
+                        referenced_blocks,
+                        None,
                         tmpl.initial_auto_escape()
                     );
                 }
                 Instruction::PushAutoEscape => {
                     let value = stack.pop();
-                    auto_escape_stack.push(auto_escape);
-                    auto_escape = match value.as_primitive() {
+                    auto_escape_stack.push(state.auto_escape);
+                    state.auto_escape = match value.as_primitive() {
                         Some(Primitive::Str("html")) => AutoEscape::Html,
                         Some(Primitive::Str("none")) | Some(Primitive::Bool(false)) => {
                             AutoEscape::None
@@ -579,7 +681,7 @@ impl<'env> Vm<'env> {
                     };
                 }
                 Instruction::PopAutoEscape => {
-                    auto_escape = auto_escape_stack.pop().unwrap();
+                    state.auto_escape = auto_escape_stack.pop().unwrap();
                 }
                 Instruction::BeginCapture => {
                     begin_capture!();
@@ -590,21 +692,19 @@ impl<'env> Vm<'env> {
                 Instruction::ApplyFilter(name) => {
                     let args = try_ctx!(stack.pop().try_into_vec());
                     let value = stack.pop();
-                    stack.push(try_ctx!(self.env.apply_filter(name, value, args)));
+                    stack.push(try_ctx!(state.apply_filter(name, value, args)));
                 }
                 Instruction::PerformTest(name) => {
                     let args = try_ctx!(stack.pop().try_into_vec());
                     let value = stack.pop();
-                    stack.push(Value::from(try_ctx!(self
-                        .env
-                        .perform_test(name, value, args))));
+                    stack.push(Value::from(try_ctx!(state.perform_test(name, value, args))));
                 }
                 Instruction::CallFunction(function_name) => {
                     let args = try_ctx!(stack.pop().try_into_vec());
                     // super is a special function reserved for super-ing into blocks.
                     if *function_name == "super" {
                         let mut inner_blocks = blocks.clone();
-                        let name = match block_stack.last() {
+                        let name = match state.current_block {
                             Some(name) => name,
                             None => {
                                 bail!(Error::new(
@@ -622,8 +722,8 @@ impl<'env> Vm<'env> {
                         } else {
                             panic!("attempted to super unreferenced block");
                         }
-                    } else if let Some(func) = context.lookup(self.env, function_name) {
-                        stack.push(try_ctx!(func.call(self.env, args)));
+                    } else if let Some(func) = state.ctx.load(self.env, function_name) {
+                        stack.push(try_ctx!(func.call(state, args)));
                     } else {
                         bail!(Error::new(
                             ErrorKind::ImpossibleOperation,
@@ -634,12 +734,12 @@ impl<'env> Vm<'env> {
                 Instruction::CallMethod(name) => {
                     let args = try_ctx!(stack.pop().try_into_vec());
                     let obj = stack.pop();
-                    stack.push(try_ctx!(obj.call_method(self.env, name, args)));
+                    stack.push(try_ctx!(obj.call_method(state, name, args)));
                 }
                 Instruction::CallObject => {
                     let args = try_ctx!(stack.pop().try_into_vec());
                     let obj = stack.pop();
-                    stack.push(try_ctx!(obj.call(self.env, args)));
+                    stack.push(try_ctx!(obj.call(state, args)));
                 }
                 Instruction::DupTop => {
                     stack.push(stack.peek().clone());
@@ -658,13 +758,14 @@ impl<'env> Vm<'env> {
 
 /// Simple version of eval without environment or vm.
 #[cfg(feature = "unstable_machinery")]
-pub fn simple_eval<S: Serialize>(
+pub fn simple_eval<S: serde::Serialize>(
     instructions: &Instructions<'_>,
-    root: S,
+    ctx: S,
     output: &mut String,
 ) -> Result<Option<Value>, Error> {
     let env = Environment::new();
     let empty_blocks = BTreeMap::new();
     let vm = Vm::new(&env);
+    let root = Value::from_serializable(&ctx);
     vm.eval(instructions, root, &empty_blocks, AutoEscape::None, output)
 }
